#include "config.h"
#include "ui_config.h"
#include "netdiscovery.h"
#include "ftpserver.h"
#include <QSettings>
#include <QHostInfo>
#include <QFileDialog>
#include <QTranslator>
#include "sharedfolders.h"
#include "globals.h"
#include <QCryptographicHash>
#include <QDesktopServices>
#include <QUrl>
#include <QProcess>
#include <QClipboard>
#include "bonjourregistrar.h"

using namespace qshare;

Config *Config::pInstance = NULL;

QMutex Config::sMutex(QMutex::Recursive);

Config *Config::instance()
{
	if (pInstance)
		return pInstance;
	return new Config;
}

void Config::deleteInstance()
{
	if (pInstance)
		pInstance->deleteLater();
}

Config::Config(QWidget *parent) :
    QDialog(parent),
    m_ui(new Ui::Config)
{
	pInstance = this;
	m_ui->setupUi(this);

	connect(m_ui->anonymous, SIGNAL(toggled(bool)), m_ui->login, SLOT(setDisabled(bool)));
	connect(m_ui->anonymous, SIGNAL(toggled(bool)), m_ui->password, SLOT(setDisabled(bool)));
	connect(m_ui->anonymous, SIGNAL(toggled(bool)), m_ui->login_label, SLOT(setDisabled(bool)));
	connect(m_ui->anonymous, SIGNAL(toggled(bool)), m_ui->password_label, SLOT(setDisabled(bool)));
	connect(m_ui->anonymous, SIGNAL(toggled(bool)), m_ui->readonlyanonymous, SLOT(setDisabled(bool)));
	connect(m_ui->anonymous, SIGNAL(toggled(bool)), m_ui->roa_label, SLOT(setDisabled(bool)));
	connect(m_ui->bAdd, SIGNAL(clicked()), this, SLOT(addFolder()));
	connect(m_ui->bRemove, SIGNAL(clicked()), this, SLOT(removeFolder()));

	loadSettings();

	connect(m_ui->buttonBox, SIGNAL(clicked(QAbstractButton*)), this, SLOT(processButtonClick(QAbstractButton*)));

	FtpServer::startServer();

	loadFolders();
	loadLanguages();
	connect(m_ui->sharedFolders, SIGNAL(itemChanged(QTableWidgetItem*)), this, SLOT(editFolder(QTableWidgetItem*)));
	connect(m_ui->sharedFolders, SIGNAL(itemClicked(QTableWidgetItem*)), this, SLOT(editPath(QTableWidgetItem*)));
	connect(m_ui->translation, SIGNAL(currentIndexChanged(int)), this, SLOT(changeLanguage()));

	setWindowFlags(Qt::Dialog | Qt::CustomizeWindowHint | Qt::WindowTitleHint);
    setWindowTitle(windowTitle() + ' ' + tr("on") + ' ' + Config::instance()->getHostname());
}

Config::~Config()
{
    delete m_ui;
	pInstance = NULL;
}

void Config::processButtonClick(QAbstractButton *b)
{
	switch (m_ui->buttonBox->standardButton(b))
	{
	case QDialogButtonBox::Ok:
		saveSettings();
		loadFolders();
		break;
	case QDialogButtonBox::Apply:
		saveSettings();
		loadFolders();
		break;
	case QDialogButtonBox::Cancel:
		restoreFolders();
		loadSettings();
		break;
	default:
		break;
	}
}

void Config::changeEvent(QEvent *e)
{
    QDialog::changeEvent(e);
    switch (e->type()) {
    case QEvent::LanguageChange:
		saveSettings();
        m_ui->retranslateUi(this);
        loadLanguages();
		loadSettings();
        setWindowTitle(windowTitle() + ' ' + tr("on") + ' ' + Config::instance()->getHostname());
        break;
    default:
        break;
    }
}

bool Config::isZeroconfEnabled()
{
	QMutexLocker mLock(&sMutex);
	settings->beginGroup("config");
	const bool zeroconf = settings->value("zeroconf", true).toBool();
	settings->endGroup();
	return zeroconf;
}

void Config::setZeroconfEnabled(bool enabled)
{
	QMutexLocker mLock(&sMutex);
	settings->beginGroup("config");
	settings->setValue("zeroconf", enabled);
	settings->endGroup();
}

bool Config::isAnonymous()
{
	QMutexLocker mLock(&sMutex);
	settings->beginGroup("config");
	const bool anonymous = settings->value("anonymous", false).toBool();
	settings->endGroup();
	return anonymous;
}

bool Config::isAnonymousReadAllowed()
{
	QMutexLocker mLock(&sMutex);
	settings->beginGroup("config");
	const bool readonlyanonymous = settings->value("readonlyanonymous", false).toBool();
	settings->endGroup();
	return readonlyanonymous;
}

bool Config::showHiddenFiles()
{
	QMutexLocker mLock(&sMutex);
	settings->beginGroup("config");
	const bool hidden = settings->value("hidden", false).toBool();
	settings->endGroup();
	return hidden;
}

bool Config::checkPassword(const QString &passwd)
{
	QMutexLocker mLock(&sMutex);
	settings->beginGroup("config");
	const QByteArray password = settings->value("password").toByteArray();
	const bool isEncrypted = settings->value("encrypted", false).toBool();
	settings->endGroup();
	if (isEncrypted)
		return password == QCryptographicHash::hash(passwd.toUtf8(), QCryptographicHash::Sha1).toHex();
	else
		return password == passwd.toUtf8();
}

QString Config::getLogin()
{
	QMutexLocker mLock(&sMutex);
	settings->beginGroup("config");
	const QString login = settings->value("login").toString();
	settings->endGroup();
	return login;
}

bool Config::checkID(const UserID &uid, const bool read)
{
	if (isAnonymous())
		return true;
	if (read && isAnonymousReadAllowed())
		return true;
	return getUID() == uid;
}

UserID Config::getUID()
{
	if (isAnonymous())
		return UserID("anonymous", QString());

	QMutexLocker mLock(&sMutex);
	settings->beginGroup("config");
	QByteArray password = settings->value("password").toByteArray();
	const bool isEncrypted = settings->value("encrypted", false).toBool();
	settings->endGroup();
	if (!isEncrypted)
		password == QCryptographicHash::hash(password, QCryptographicHash::Sha1).toHex();
	return UserID(getLogin(), password);
}

QString Config::getHostname()
{
	return QHostInfo::localHostName();
}

int Config::getServerPort()
{
	return 2000;
}

void Config::updateServerStatus()
{
	QMutexLocker mLock(&sMutex);
	FtpServer::restartServer();
}

void Config::saveSettings()
{
	QMutexLocker mLock(&sMutex);
	bool bServerNeedsToBeRestarted = false;
	const bool b_isAnonymous = isAnonymous();
	const bool b_isAnonymousReadAllowed = isAnonymousReadAllowed();
	const bool b_isZeroconfEnabled = isZeroconfEnabled();
	const QString login = getLogin();
	settings->beginGroup("config");
	if (b_isAnonymous != m_ui->anonymous->isChecked())
	{
		settings->setValue("anonymous", m_ui->anonymous->isChecked());
		bServerNeedsToBeRestarted = true;
	}
	if (b_isAnonymousReadAllowed != m_ui->readonlyanonymous->isChecked())
	{
		settings->setValue("readonlyanonymous", m_ui->readonlyanonymous->isChecked());
		bServerNeedsToBeRestarted = true;
	}
	if (m_ui->password->text().toUtf8() != settings->value("password", QByteArray()).toByteArray())
	{
		settings->setValue("password", QCryptographicHash::hash(m_ui->password->text().toUtf8(), QCryptographicHash::Sha1).toHex());
		settings->setValue("encrypted", true);
		bServerNeedsToBeRestarted = true;
	}
	if (login != m_ui->login->text())
	{
		settings->setValue("login", m_ui->login->text());
		bServerNeedsToBeRestarted = true;
	}
	bool bNetDiscoveryNeedsToBeRestarted = false;
	if (b_isZeroconfEnabled != m_ui->zeroconf->isChecked())
	{
		settings->setValue("zeroconf", m_ui->zeroconf->isChecked());
		bNetDiscoveryNeedsToBeRestarted = true;
	}
	settings->setValue("splash", m_ui->cbSplash->isChecked());
	settings->setValue("hidden", m_ui->hidden->isChecked());
	settings->setValue("bUseCustomCommand", m_ui->gCustom->isChecked());
	settings->setValue("sCustomCommand", m_ui->command->text());
	settings->endGroup();
	if (m_ui->gCustom->isChecked())
		QDesktopServices::setUrlHandler("ftp", this, "openURL");
	else
		QDesktopServices::unsetUrlHandler("ftp");
	settings->sync();
	if (bNetDiscoveryNeedsToBeRestarted)
		NetDiscovery::instance()->restart();
	if (bServerNeedsToBeRestarted)
		updateServerStatus();
}

void Config::loadSettings()
{
	QMutexLocker mLock(&sMutex);
	settings->beginGroup("config");
	m_ui->anonymous->setChecked(settings->value("anonymous",true).toBool());
	m_ui->readonlyanonymous->setChecked(settings->value("readonlyanonymous",false).toBool());
	m_ui->password->setText(QString::fromUtf8(settings->value("password",QByteArray()).toByteArray()));
	if (!settings->value("encrypted", false).toBool())
	{
		settings->setValue("password", QCryptographicHash::hash(m_ui->password->text().toUtf8(), QCryptographicHash::Sha1).toHex());
		settings->setValue("encrypted", true);
		m_ui->password->setText(QString::fromUtf8(settings->value("password",QByteArray()).toByteArray()));
	}
	m_ui->login->setText(settings->value("login",qshare::localUserName).toString());
	m_ui->cbSplash->setChecked(settings->value("splash", true).toBool());
	m_ui->hidden->setChecked(settings->value("hidden",false).toBool());
	m_ui->gCustom->setChecked(settings->value("bUseCustomCommand", false).toBool());
	m_ui->command->setText(settings->value("sCustomCommand", "firefox \"%U\"").toString());
	if (m_ui->gCustom->isChecked())
		QDesktopServices::setUrlHandler("ftp", this, "openURL");
	else
		QDesktopServices::unsetUrlHandler("ftp");
	m_ui->zeroconf->setChecked(settings->value("zeroconf",true).toBool());
#ifdef USE_ZEROCONF
	m_ui->zeroconf->setEnabled(true);
	m_ui->lbl_zeroconf->setEnabled(true);
#else
	m_ui->zeroconf->setEnabled(false);
	m_ui->lbl_zeroconf->setEnabled(false);
#endif
	settings->endGroup();
}

void Config::addFolder()
{
	QMutexLocker mLock(&sMutex);
	int newIdx = m_ui->sharedFolders->rowCount();
	m_ui->sharedFolders->setRowCount(newIdx + 1);
	QString name = tr("new folder #") + QString::number(newIdx);
	folderNames << name;
	QTableWidgetItem *item;
	m_ui->sharedFolders->setItem(newIdx, 0, item = new QTableWidgetItem(QIcon(":/icons/computer.png"), name));
	item->setFlags(Qt::ItemIsEditable | Qt::ItemIsEnabled | Qt::ItemIsSelectable);

	m_ui->sharedFolders->setItem(newIdx, 1, item = new QTableWidgetItem(tr("forbidden")));
	item->setFlags(Qt::ItemIsUserCheckable | Qt::ItemIsEnabled | Qt::ItemIsSelectable);
	item->setCheckState(Qt::Unchecked);

	m_ui->sharedFolders->setItem(newIdx, 2, item = new QTableWidgetItem(""));
	item->setFlags(Qt::ItemIsEditable | Qt::ItemIsEnabled | Qt::ItemIsSelectable);

	SharedFolders::instance()->addFolder(name, SharedFolders::Folder("",false));
}

void Config::removeFolder()
{
	QMutexLocker mLock(&sMutex);
	QList<QTableWidgetItem*> list = m_ui->sharedFolders->selectedItems();
	if (list.isEmpty())
		return;
	SharedFolders::instance()->removeFolder(folderNames[list.front()->row()]);
	QMap<QString, SharedFolders::Folder> foldersBackup = this->foldersBackup;
	loadFolders();
	this->foldersBackup = foldersBackup;
}

void Config::editFolder(QTableWidgetItem *item)
{
	if (item == NULL || item->row() == -1)
		return;

	QMutexLocker mLock(&sMutex);

	QMap<QString, SharedFolders::Folder> folders = SharedFolders::instance()->getFolderList();
	QString name = folderNames[ item->row() ];
	SharedFolders::Folder folder = folders[ name ];
	SharedFolders::instance()->removeFolder( name );
	switch (item->column())
	{
	case 0:
		name = item->text();
		folderNames[item->row()] = name;
		break;
	case 1:
		folder.setWritable(item->checkState() == Qt::Checked);
		item->setText(folder.isWritable() ? tr("allowed") : tr("forbidden"));
		break;
	case 2:
		folder.setPath(item->text());
		break;
	};
	SharedFolders::instance()->addFolder(name, folder);
}

void Config::editPath(QTableWidgetItem *item)
{
	if (item == NULL || item->row() == -1)
		return;

	QMutexLocker mLock(&sMutex);

	QMap<QString, SharedFolders::Folder> folders = SharedFolders::instance()->getFolderList();
	QString name = folderNames[ item->row() ];
	SharedFolders::Folder folder = folders[ name ];
	SharedFolders::instance()->removeFolder( name );
	switch (item->column())
	{
	case 2:
		{
			const QString path = QFileDialog::getExistingDirectory(this, tr("Select a folder"), QString(), QFileDialog::ShowDirsOnly | QFileDialog::DontUseNativeDialog);
			item->setText(path);
			folder.setPath(path);
		}
		break;
	};
	SharedFolders::instance()->addFolder(name, folder);
}

void Config::restoreFolders()
{
	QMutexLocker mLock(&sMutex);

	QMap<QString, SharedFolders::Folder> folders = SharedFolders::instance()->getFolderList();
	for(QMap<QString, SharedFolders::Folder>::const_iterator it = folders.begin() ; it != folders.end() ; ++it)
		SharedFolders::instance()->removeFolder(it.key());
	for(QMap<QString, SharedFolders::Folder>::const_iterator it = foldersBackup.begin() ; it != foldersBackup.end() ; ++it)
		SharedFolders::instance()->addFolder(it.key(), it.value());
	loadFolders();
}

void Config::loadFolders()
{
	QMutexLocker mLock(&sMutex);
	m_ui->sharedFolders->setRowCount(SharedFolders::instance()->nbFolders());
	folderNames.clear();
	QMap<QString, SharedFolders::Folder> folders = SharedFolders::instance()->getFolderList();
	foldersBackup = folders;
	int n = 0;
	for(QMap<QString, SharedFolders::Folder>::iterator i = folders.begin() ; i != folders.end() ; ++i)
	{
		folderNames << i.key();
		QTableWidgetItem *item;
		m_ui->sharedFolders->setItem(n, 0, item = new QTableWidgetItem(QIcon(":/icons/computer.png"), i.key()));
		item->setFlags(Qt::ItemIsEditable | Qt::ItemIsEnabled | Qt::ItemIsSelectable);

		m_ui->sharedFolders->setItem(n, 1, item = new QTableWidgetItem(i.value().isWritable() ? tr("allowed") : tr("forbidden")));
		item->setFlags(Qt::ItemIsUserCheckable | Qt::ItemIsEnabled | Qt::ItemIsSelectable);
		item->setCheckState(i.value().isWritable() ? Qt::Checked : Qt::Unchecked);

		m_ui->sharedFolders->setItem(n, 2, item = new QTableWidgetItem(i.value().getPath()));
		item->setFlags(Qt::ItemIsEditable | Qt::ItemIsEnabled | Qt::ItemIsSelectable);
		++n;
	}
}

void Config::addPath(const bool writable, const QString &path)
{
	// Since we can't share it if it doesn't exist
	// exclude non-existing folders
	if (!QDir(path).exists())
		return;

	// Check if it's already there
	const QMap<QString, SharedFolders::Folder> folders = SharedFolders::instance()->getFolderList();
	for(QMap<QString, SharedFolders::Folder>::const_iterator it = folders.begin(), end = folders.end() ; it != end ; ++it)
		if (it.value().getPath() == path				// No need to share it twice
			&& it.value().isWritable() == writable)		// with the same access rights
			return;

	QStringList nodes = QDir::convertSeparators(path).split('/', QString::SkipEmptyParts);
	if (nodes.empty())		// Don't share '/' (anyway how could we do that automatically without knowning the folder name ?)
		return;
	SharedFolders::instance()->addFolder(nodes.back(), SharedFolders::Folder(path, writable));
	// Add it to the backup list (we don't want to cancel it when the cancel button is pressed)
	sMutex.lock();
	foldersBackup[nodes.back()] = SharedFolders::Folder(path, writable);
	sMutex.unlock();
	loadFolders();

	// Set clipboard content to qShare's URL to the folder
	qApp->clipboard()->setText("ftp://" + QHostInfo::localHostName() + ":2000/" + qshare::localUserName + '/' + nodes.back());
}

void Config::removePath(const QString &path)
{
	sMutex.lock();
	const QMap<QString, SharedFolders::Folder> folders = SharedFolders::instance()->getFolderList();
	for(QMap<QString, SharedFolders::Folder>::const_iterator it = folders.begin(), end = folders.end() ; it != end ; ++it)
		if (it.value().getPath() == path)
		{
			SharedFolders::instance()->removeFolder(it.key());
			// Remove it from the backup list (we don't want to reinsert it back when the cancel button is pressed)
			foldersBackup.remove(it.key());
		}
	sMutex.unlock();
	loadFolders();
}

void Config::changeLanguage()
{
	QMutexLocker mLock(&sMutex);
	QString code = m_ui->translation->itemData(m_ui->translation->currentIndex()).toString();
	settings->setValue("locale", code);
	qAppTranslator->load(QString(I18N_PATH) + "qshare_" + code + ".qm");
#ifdef SYSTEM_I18N_PATH
	qQtTranslator->load(QString(SYSTEM_I18N_PATH) + "qt_" + code + ".qm");
#else
	qQtTranslator->load(QString(I18N_PATH) + "qt_" + code + ".qm");
#endif
}

void Config::loadLanguages()
{
	disconnect(m_ui->translation, SIGNAL(currentIndexChanged(int)), this, SLOT(changeLanguage()));
	int idx = m_ui->translation->currentIndex();
	m_ui->translation->clear();
    QDir dir(QString(I18N_PATH), "qshare_*.qm");
	foreach(QString filename, dir.entryList())
	{
		QTranslator translator;
		translator.load(QString(I18N_PATH) + filename);
		m_ui->translation->addItem(translator.translate("options", "LANG"), filename.mid(7, 2));
	}
	if (idx == -1)
	{
		QString code = settings->value("locale", QLocale::system().name().left(2)).toString();
		for(int i = 0 ; idx == -1 && i < m_ui->translation->count() ; ++i)
			if (m_ui->translation->itemData(i) == code)
				idx = i;
	}
	m_ui->translation->setCurrentIndex(idx);
	connect(m_ui->translation, SIGNAL(currentIndexChanged(int)), this, SLOT(changeLanguage()));
}

void Config::openURL(const QUrl &url)
{
	QMutexLocker mLock(&sMutex);
	settings->beginGroup("config");
	const QString customCommand = settings->value("sCustomCommand", "firefox \"%U\"").toString();
	settings->endGroup();
	QString command = customCommand;
	command.replace("%U", url.toString(), Qt::CaseInsensitive);
	QProcess::startDetached(command);
}

void Config::lock()
{
	sMutex.lock();
}

void Config::unlock()
{
	sMutex.unlock();
}
