#include "ipfileserver.h"
#include <QTime>
#include <QCoreApplication>
#include "sharedfolders.h"
#include <QDebug>
#include "userid.h"
#include "config.h"

IPFileServer *IPFileServer::pInstance = NULL;
IPFileServer *IPFileServer::instance()
{
	if (!pInstance)
		pInstance = new IPFileServer;
	return pInstance;
}

IPFileServer::IPFileServer() : IPObject("IPFileServer", 6845), nextTransactionID(0), fileID(0)
{
}

IPFileServer::~IPFileServer()
{
}

QByteArray IPFileServer::waitForID(const quint32 ID)
{
	QByteArray result;
	QTime timer;
	timer.start();
	while(!messageQueue.contains(ID))
	{
		waitForData(10);
		if (timer.elapsed() >= 10000)			// Timeout
		{
			messageQueue.remove(ID);
			return result;
		}
	}
	result = messageQueue[ID];
	messageQueue.remove(ID);
	return result;
}

void IPFileServer::receive(const QByteArray &message)
{
	QDataStream msg(message);
	QString type;
	msg >> type;
	if (type == "answer")
	{
		quint32 ID;
		QString sender;
		msg >> sender >> ID;
		if (sender == qshare::localUserName)
			messageQueue[ID] = message.right(message.length() - msg.device()->pos());
		return;
	}
	if (type == "seek")
	{
		quint32 ID;
		QString sender, user;
		quint32 fID;
		qint64 offset;
		msg >> sender >> ID >> user >> fID >> offset;
		if (user != qshare::localUserName)
			return;
		send("answer", sender, ID, seek(user, fID, offset));
		return;
	}
	if (type == "open")
	{
		quint32 ID;
		QString sender, user, filename;
		int mode;
		UserID uid;
		msg >> sender >> ID >> uid >> user >> filename >> mode;
		if (user != qshare::localUserName)
			return;
		send("answer", sender, ID, open(uid, user, filename, QFile::OpenMode(mode)));
		return;
	}
	if (type == "close")
	{
		quint32 ID;
		QString sender, user;
		quint32 fID;
		msg >> sender >> ID >> user >> fID;
		if (user != qshare::localUserName)
			return;
		close(user, fID);
		send("answer", sender, ID);
		return;
	}
	if (type == "read")
	{
		quint32 ID;
		QString sender, user;
		quint32 fID;
		qint64 maxlen;
		msg >> sender >> ID >> user >> fID >> maxlen;
		if (user != qshare::localUserName)
			return;
		send("answer", sender, ID, read(user, fID, maxlen));
		return;
	}
	if (type == "write")
	{
		quint32 ID;
		QString sender, user;
		quint32 fID;
		QByteArray data;
		msg >> sender >> ID >> user >> fID >> data;
		if (user != qshare::localUserName)
			return;
		write(user, fID, data);
		send("answer", sender, ID);
		return;
	}
	if (type == "flush")
	{
		quint32 ID;
		QString sender, user;
		quint32 fID;
		msg >> sender >> ID >> user >> fID;
		if (user != qshare::localUserName)
			return;
		flush(user, fID);
		send("answer", sender, ID);
		return;
	}
	if (type == "isReadable")
	{
		quint32 ID;
		QString sender, user;
		quint32 fID;
		msg >> sender >> ID >> user >> fID;
		if (user != qshare::localUserName)
			return;
		send("answer", sender, ID, isReadable(user, fID));
		return;
	}
	if (type == "isWritable")
	{
		quint32 ID;
		QString sender, user;
		quint32 fID;
		msg >> sender >> ID >> user >> fID;
		if (user != qshare::localUserName)
			return;
		send("answer", sender, ID, isWritable(user, fID));
		return;
	}
}

bool IPFileServer::seek(const QString &user, quint32 ID, qint64 offset)
{
	if (user != qshare::localUserName)
	{
		QByteArray result = remote("seek", user, ID, offset);
		QDataStream reader(result);
		bool b;
		reader >> b;
		return b;
	}

	if (!files.contains(ID))
		return true;
	return files[ID]->seek(offset);
}

bool IPFileServer::isReadable(const QString &user, quint32 ID)
{
	if (user != qshare::localUserName)
	{
		QByteArray result = remote("isReadable", user, ID);
		QDataStream reader(result);
		bool b;
		reader >> b;
		return b;
	}

	if (!files.contains(ID))
		return false;
	return files[ID]->isReadable();
}

bool IPFileServer::isWritable(const QString &user, quint32 ID)
{
	if (user != qshare::localUserName)
	{
		QByteArray result = remote("isWritable", user, ID);
		QDataStream reader(result);
		bool b;
		reader >> b;
		return b;
	}

	if (!files.contains(ID))
		return false;
	return files[ID]->isWritable();
}

quint32 IPFileServer::open(const UserID &uid, const QString &user, const QString &filename, QFile::OpenMode mode)
{
	if (user != qshare::localUserName)
	{
		QByteArray result = remote("open", uid, user, filename, int(mode));
		QDataStream reader(result);
		quint32 ID;
		reader >> ID;
		return ID;
	}

	if (!Config::checkID(uid, true))
		return 0U;
	if (!Config::checkID(uid) && mode & QFile::WriteOnly)
		return 0U;

	++fileID;
	quint32 ID = fileID;

	const QString realFilename = SharedFolders::instance()->getRealPath(filename);
	files[ID] = new QFile(realFilename, this);
	files[ID]->open(mode);

	return ID;
}

void IPFileServer::close(const QString &user, quint32 ID)
{
	if (user != qshare::localUserName)
	{
		remote("close", user, ID);
		return;
	}

	if (!files.contains(ID))
		return;
	files[ID]->close();
	delete files[ID];
	files.remove(ID);
}

void IPFileServer::flush(const QString &user, quint32 ID)
{
	if (user != qshare::localUserName)
	{
		remote("flush", user, ID);
		return;
	}

	if (!files.contains(ID))
		return;
	files[ID]->flush();
}

QByteArray IPFileServer::read(const QString &user, quint32 ID, qint64 maxlen)
{
	if (user != qshare::localUserName)
	{
		QByteArray result = remote("read", user, ID, maxlen);
		QDataStream reader(result);
		QByteArray data;
		reader >> data;
		return data;
	}

	if (!files.contains(ID))
		return QByteArray();
	return files[ID]->read(maxlen);
}

void IPFileServer::write(const QString &user, quint32 ID, const QByteArray &data)
{
	if (user != qshare::localUserName)
	{
		remote("write", user, ID, data);
		return;
	}

	if (!files.contains(ID))
		return;
	files[ID]->write(data);
}
